@@ -32,22 +32,27 @@
 # include <config.h>
 #endif
 
+#if !defined(_LARGEFILE_SOURCE)
+# define _LARGEFILE_SOURCE
+#endif
+#if !defined(_LARGEFILE64_SOURCE)
+# define _LARGEFILE64_SOURCE
+#endif
+#if !defined(_FILE_OFFSET_BITS)
+# define _FILE_OFFSET_BITS 64
+#endif
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
 #include <math.h>
 
+#include "stack_alloc.h"
+
 #ifdef WIN32
-# include <malloc.h>
 # include <windows.h> /*GetFileType()*/
 # include <io.h>      /*_get_osfhandle()*/
-#else
-# ifdef HAVE_ALLOCA_H
-#  include <alloca.h>
-# else
-#  include <stdlib.h>
-# endif
 #endif
 
 #ifdef ENABLE_NLS
@@ -67,6 +72,8 @@
 #include <ogg/ogg.h>
 #include "opusenc.h"
 #include "speex_resampler.h"
+#include "lpc.h"
+#include "opus_header.h"
 
 #ifdef HAVE_LIBFLAC
 #include "flac.h"
@@ -288,8 +295,9 @@ int aiff_open(FILE *in, oe_enc_opt *opt, unsigned char *buf, int buflen)
     unsigned int len;
     unsigned char *buffer;
     unsigned char buf2[8];
+    int bigendian = 1;
     aiff_fmt format;
-    aifffile *aiff = malloc(sizeof(aifffile));
+    aifffile *aiff;
     int i;
     (void)buflen;/*unused*/
 
@@ -323,8 +331,6 @@ int aiff_open(FILE *in, oe_enc_opt *opt, unsigned char *buf, int buflen)
     format.samplesize = READ_U16_BE(buffer+6);
     format.rate = (int)read_IEEE80(buffer+8);
 
-    aiff->bigendian = 1;
-
     if(aifc)
     {
         if(len < 22)
@@ -335,11 +341,11 @@ int aiff_open(FILE *in, oe_enc_opt *opt, unsigned char *buf, int buflen)
 
         if(!memcmp(buffer+18, "NONE", 4))
         {
-            aiff->bigendian = 1;
+            bigendian = 1;
         }
         else if(!memcmp(buffer+18, "sowt", 4))
         {
-            aiff->bigendian = 0;
+            bigendian = 0;
         }
         else
         {
@@ -379,11 +385,13 @@ int aiff_open(FILE *in, oe_enc_opt *opt, unsigned char *buf, int buflen)
         opt->read_samples = wav_read; /* Similar enough, so we use the same */
         opt->total_samples_per_channel = format.totalframes;
 
+        aiff = malloc(sizeof(aifffile));
         aiff->f = in;
         aiff->samplesread = 0;
         aiff->channels = format.channels;
         aiff->samplesize = format.samplesize;
         aiff->totalsamples = format.totalframes;
+        aiff->bigendian = bigendian;
 
         if(aiff->channels>3)
           fprintf(stderr,"WARNING: AIFF[-C] files with greater than three channels use\n"
@@ -429,25 +437,13 @@ int wav_id(unsigned char *buf, int len)
     return 1;
 }
 
-static int wav_permute_matrix[8][8] =
-{
-  {0},              /* 1.0 mono   */
-  {0,1},            /* 2.0 stereo */
-  {0,2,1},          /* 3.0 channel ('wide') stereo */
-  {0,1,2,3},        /* 4.0 discrete quadraphonic */
-  {0,2,1,3,4},      /* 5.0 surround */
-  {0,2,1,4,5,3},    /* 5.1 surround */
-  {0,2,1,5,6,4,3},  /* 6.1 surround */
-  {0,2,1,6,7,4,5,3} /* 7.1 surround (classic theater 8-track) */
-};
-
 int wav_open(FILE *in, oe_enc_opt *opt, unsigned char *oldbuf, int buflen)
 {
     unsigned char buf[40];
     unsigned int len;
     int samplesize;
     wav_fmt format;
-    wavfile *wav = malloc(sizeof(wavfile));
+    wavfile *wav;
     int i;
     (void)buflen;/*unused*/
     (void)oldbuf;/*unused*/
@@ -579,6 +575,7 @@ int wav_open(FILE *in, oe_enc_opt *opt, unsigned char *oldbuf, int buflen)
         opt->rate = format.samplerate;
         opt->channels = format.channels;
 
+        wav = malloc(sizeof(wavfile));
         wav->f = in;
         wav->samplesread = 0;
         wav->bigendian = 0;
@@ -609,8 +606,21 @@ int wav_open(FILE *in, oe_enc_opt *opt, unsigned char *oldbuf, int buflen)
             }
             else
             {
-                opt->total_samples_per_channel = (ftell(in) - pos)/
-                    (format.channels*samplesize);
+#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
+                opt->total_samples_per_channel = _ftelli64(in);
+#elif defined HAVE_FSEEKO
+                opt->total_samples_per_channel = ftello(in);
+#else
+                opt->total_samples_per_channel = ftell(in);
+#endif
+                if(opt->total_samples_per_channel>pos)
+                {
+                   opt->total_samples_per_channel = (opt->total_samples_per_channel-pos)/(format.channels*samplesize);
+                }
+                else
+                {
+                   opt->total_samples_per_channel=0;
+                }
                 fseek(in,pos, SEEK_SET);
             }
         }
@@ -820,22 +830,43 @@ typedef struct {
     void *real_readdata;
     ogg_int64_t *original_samples;
     int channels;
+    int lpc_ptr;
     int *extra_samples;
+    float *lpc_out;
 } padder;
 
+/* Read audio data, appending padding to make up any gap
+ * between the available and requested number of samples
+ * with LPC-predicted data to minimize the pertubation of
+ * the valid data that falls in the same frame.
+ */
 static long read_padder(void *data, float *buffer, int samples) {
     padder *d = data;
     long in_samples = d->real_reader(d->real_readdata, buffer, samples);
     int i, extra=0;
+    const int lpc_order=32;
 
     if(d->original_samples)*d->original_samples+=in_samples;
 
     if(in_samples<samples){
+      if(d->lpc_ptr<0){
+        d->lpc_out=calloc(d->channels * *d->extra_samples, sizeof(*d->lpc_out));
+        if(in_samples>lpc_order*2){
+          float *lpc=alloca(lpc_order*sizeof(*lpc));
+          for(i=0;i<d->channels;i++){
+            vorbis_lpc_from_data(buffer+i,lpc,in_samples,lpc_order,d->channels);
+            vorbis_lpc_predict(lpc,buffer+i+(in_samples-lpc_order)*d->channels,
+                               lpc_order,d->lpc_out+i,*d->extra_samples,d->channels);
+          }
+        }
+        d->lpc_ptr=0;
+      }
       extra=samples-in_samples;
       if(extra>*d->extra_samples)extra=*d->extra_samples;
       *d->extra_samples-=extra;
     }
-    for(i=0;i<extra*d->channels;i++)buffer[(in_samples*d->channels)+i]=0;
+    memcpy(buffer+in_samples*d->channels,d->lpc_out+d->lpc_ptr*d->channels,extra*d->channels*sizeof(*buffer));
+    d->lpc_ptr+=extra;
     return in_samples+extra;
 }
 
@@ -850,6 +881,8 @@ void setup_padder(oe_enc_opt *opt,ogg_int64_t *original_samples) {
     d->channels = opt->channels;
     d->extra_samples = &opt->extraout;
     d->original_samples=original_samples;
+    d->lpc_ptr = -1;
+    d->lpc_out = NULL;
 }
 
 void clear_padder(oe_enc_opt *opt) {
@@ -858,6 +891,7 @@ void clear_padder(oe_enc_opt *opt) {
     opt->read_samples = d->real_reader;
     opt->readdata = d->real_readdata;
 
+    if(d->lpc_out)free(d->lpc_out);
     free(d);
 }
 
@@ -884,17 +918,18 @@ static long read_resampled(void *d, float *buffer, int samples)
       int i;
       int reading, ret;
       unsigned in_len, out_len;
-      reading=1024-*inbuf;
+      out_len=samples-out_samples;
+      reading=rs->bufsize-*inbuf;
+      if(reading>1024)reading=1024;
       ret=rs->real_reader(rs->real_readdata, pcmbuf+*inbuf*rs->channels, reading);
       *inbuf+=ret;
       in_len=*inbuf;
-      out_len=samples-out_samples;
       speex_resampler_process_interleaved_float(rs->resampler, pcmbuf, &in_len, buffer+out_samples*rs->channels, &out_len);
+      out_samples+=out_len;
       if(ret==0&&in_len==0){
         for(i=out_samples*rs->channels;i<samples*rs->channels;i++)buffer[i]=0;
         return out_samples;
       }
-      out_samples+=out_len;
       for(i=0;i<rs->channels*(*inbuf-(long int)in_len);i++)pcmbuf[i]=pcmbuf[i+rs->channels*in_len];
       *inbuf-=in_len;
     }
@@ -905,7 +940,7 @@ int setup_resample(oe_enc_opt *opt, int complexity, long outfreq) {
     resampler *rs = calloc(1, sizeof(resampler));
     int err;
 
-    rs->bufsize = 2048; /* Shrug */
+    rs->bufsize = 5760*2; /* Have at least two output frames worth, just in case of ugly ratios */
     rs->bufpos = 0;
 
     rs->real_reader = opt->read_samples;
@@ -983,15 +1018,16 @@ int setup_downmix(oe_enc_opt *opt, int out_channels) {
       /*8*/  {{1,0},{0.7071f,0.7071f},{0,1},{0.866f,0.5f},{0.5f,0.866f},{0.866f,0.5f},{0.5f,0.866f},{0.7071f,0.7071f}},
     };
     float sum;
-    downmix *d = calloc(1, sizeof(downmix));
+    downmix *d;
     int i,j;
 
-    if(opt->channels<=out_channels || out_channels>2 || (out_channels==2&&opt->channels>8)) {
+    if(opt->channels<=out_channels || out_channels>2 || (out_channels==2&&opt->channels>8) || opt->channels<=0 || out_channels<=0) {
         fprintf(stderr, "Downmix must actually downmix and only knows mono/stereo out.\n");
         if(opt->channels>8)fprintf(stderr, "Downmix also only knows how to mix >8ch to mono.\n");
         return 0;
     }
 
+    d = calloc(1, sizeof(downmix));
     d->bufs = malloc(sizeof(float)*opt->channels*4096);
     d->matrix = malloc(sizeof(float)*opt->channels*out_channels);
     d->real_reader = opt->read_samples;
