@@ -44,47 +44,56 @@
 #include <stdlib.h>
 #include <limits.h>
 #include <string.h>
+#include <ctype.h> /*tolower()*/
 
 #include <opus.h>
 #include <opus_multistream.h>
 #include <ogg/ogg.h>
 
-#if defined WIN32 || defined _WIN32
-#include "wave_out.h"
+#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
+# include "unicode_support.h"
+# include "wave_out.h"
 /* We need the following two to set stdout to binary */
-#include <io.h>
-#include <fcntl.h>
+# include <io.h>
+# include <fcntl.h>
+# define I64FORMAT "I64d"
+#else
+# define I64FORMAT "lld"
+# define fopen_utf8(_x,_y) fopen((_x),(_y))
+# define argc_utf8 argc
+# define argv_utf8 argv
 #endif
+
 #include <math.h>
 
-#ifdef _WIN32
-#define I64FORMAT "I64d"
+#ifdef HAVE_LRINTF
+# define float2int(x) lrintf(x)
 #else
-#define I64FORMAT "lld"
+# define float2int(flt) ((int)(floor(.5+flt)))
 #endif
 
-#if defined HAVE_SYS_SOUNDCARD_H || defined HAVE_MACHINE_SOUNDCARD_H || HAVE_SOUNDCARD_H
-#ifdef HAVE_SYS_SOUNDCARD_H
-# include <sys/soundcard.h>
-#elif HAVE_MACHINE_SOUNDCARD_H
-# include <machine/soundcard.h>
-#else
-# include <soundcard.h>
-#endif
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-#include <sys/ioctl.h>
-
+#if defined HAVE_LIBSNDIO
+# include <sndio.h>
+#elif defined HAVE_SYS_SOUNDCARD_H || defined HAVE_MACHINE_SOUNDCARD_H || HAVE_SOUNDCARD_H
+# ifdef HAVE_SYS_SOUNDCARD_H
+#  include <sys/soundcard.h>
+# elif HAVE_MACHINE_SOUNDCARD_H
+#  include <machine/soundcard.h>
+# else
+#  include <soundcard.h>
+# endif
+# include <sys/types.h>
+# include <sys/stat.h>
+# include <fcntl.h>
+# include <sys/ioctl.h>
 #elif defined HAVE_SYS_AUDIOIO_H
-#include <sys/types.h>
-#include <fcntl.h>
-#include <sys/ioctl.h>
-#include <sys/audioio.h>
-#ifndef AUDIO_ENCODING_SLINEAR
-#define AUDIO_ENCODING_SLINEAR AUDIO_ENCODING_LINEAR /* Solaris */
-#endif
-
+# include <sys/types.h>
+# include <fcntl.h>
+# include <sys/ioctl.h>
+# include <sys/audioio.h>
+# ifndef AUDIO_ENCODING_SLINEAR
+#  define AUDIO_ENCODING_SLINEAR AUDIO_ENCODING_LINEAR /* Solaris */
+# endif
 #endif
 
 #include <string.h>
@@ -92,6 +101,8 @@
 #include "opus_header.h"
 #include "diag_range.h"
 #include "speex_resampler.h"
+#include "stack_alloc.h"
+#include "cpusupport.h"
 
 #define MINI(_a,_b)      ((_a)<(_b)?(_a):(_b))
 #define MAXI(_a,_b)      ((_a)>(_b)?(_a):(_b))
@@ -105,6 +116,10 @@
                            ((buf[base+1]<<8)&0xff00)| \
                            (buf[base]&0xff))
 
+#ifdef HAVE_LIBSNDIO
+struct sio_hdl *hdl;
+#endif
+
 typedef struct shapestate shapestate;
 struct shapestate {
   float * b_buf;
@@ -119,10 +134,25 @@ static inline unsigned int fast_rand(void) {
   return rngseed;
 }
 
+#ifndef HAVE_FMINF
+# define fminf(_x,_y) ((_x)<(_y)?(_x):(_y))
+#endif
+
+#ifndef HAVE_FMAXF
+# define fmaxf(_x,_y) ((_x)>(_y)?(_x):(_y))
+#endif
+
+static void quit(int _x) {
+#ifdef WIN_UNICODE
+  uninit_console_utf8();
+#endif
+  exit(_x);
+}
+
 /* This implements a 16 bit quantization with full triangular dither
    and IIR noise shaping. The noise shaping filters were designed by
    Sebastian Gesemann based on the LAME ATH curves with flattening
-   to limit their peak gain to 20dB.
+   to limit their peak gain to 20 dB.
    (Everyone elses' noise shaping filters are mildly crazy)
    The 48kHz version of this filter is just a warped version of the
    44.1kHz filter and probably could be improved by shifting the
@@ -130,7 +160,7 @@ static inline unsigned int fast_rand(void) {
    room and being more conservative against bat-ears is probably
    more important than more noise suppression.
    This process can increase the peak level of the signal (in theory
-   by the peak error of 1.5 +20dB though this much is unobservable rare)
+   by the peak error of 1.5 +20 dB though this much is unobservable rare)
    so to avoid clipping the signal is attenuated by a couple thousandths
    of a dB. Initially the approach taken here was to only attenuate by
    the 99.9th percentile, making clipping rare but not impossible (like
@@ -155,6 +185,8 @@ static inline void shape_dither_toshort(shapestate *_ss, short *_o, float *_i, i
   int mute=_ss->mute;
   b_buf=_ss->b_buf;
   a_buf=_ss->a_buf;
+  /*In order to avoid replacing digital silence with quiet dither noise
+    we mute if the output has been silent for a while*/
   if(mute>64)
     memset(a_buf,0,sizeof(float)*_CC*4);
   for(i=0;i<_n;i++)
@@ -176,9 +208,9 @@ static inline void shape_dither_toshort(shapestate *_ss, short *_o, float *_i, i
       s = s - err;
       r=(float)fast_rand()*(1/(float)UINT_MAX) - (float)fast_rand()*(1/(float)UINT_MAX);
       if (mute>16)r=0;
-      /*Clamp in float out of paranoia that the input will be >96dBFS and wrap if the
+      /*Clamp in float out of paranoia that the input will be >96 dBFS and wrap if the
         integer is clamped.*/
-      _o[pos+c] = si = lrintf(fmaxf(-32768,fminf(s + r,32767)));
+      _o[pos+c] = si = float2int(fmaxf(-32768,fminf(s + r,32767)));
       /*Including clipping in the noise shaping is generally disastrous:
         the futile effort to restore the clipped energy results in more clipping.
         However, small amounts-- at the level which could normally be created by
@@ -250,7 +282,7 @@ static void print_comments(char *comments, int length)
    }
 }
 
-FILE *out_file_open(char *outFile, int rate, int *channels)
+FILE *out_file_open(char *outFile, int *wav_format, int rate, int mapping_family, int *channels)
 {
    FILE *fout=NULL;
    /*Open output file*/
@@ -262,7 +294,7 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       if (audio_fd<0)
       {
          perror("Cannot open /dev/dsp");
-         exit(1);
+         quit(1);
       }
 
       format=AFMT_S16_NE;
@@ -270,9 +302,20 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       {
          perror("SNDCTL_DSP_SETFMT");
          close(audio_fd);
-         exit(1);
+         quit(1);
       }
 
+      if (*channels > 2)
+      {
+        /* There doesn't seem to be a way to get or set the channel
+         * matrix with the sys/soundcard api, so we can't support
+         * multichannel. We should fall back to stereo downmix.
+         */
+        fprintf(stderr, "Cannot configure multichannel playback."
+                        " Try decoding to a file instead.\n");
+        close(audio_fd);
+        quit(1);
+      }
       stereo=0;
       if (*channels==2)
          stereo=1;
@@ -280,7 +323,7 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       {
          perror("SNDCTL_DSP_STEREO");
          close(audio_fd);
-         exit(1);
+         quit(1);
       }
       if (stereo!=0)
       {
@@ -293,13 +336,39 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       {
          perror("SNDCTL_DSP_SPEED");
          close(audio_fd);
-         exit(1);
+         quit(1);
       }
       fout = fdopen(audio_fd, "w");
       if(!fout)
       {
         perror("Cannot open output");
-        exit(1);
+        quit(1);
+      }
+#elif defined HAVE_LIBSNDIO
+      struct sio_par par;
+
+      hdl = sio_open(NULL, SIO_PLAY, 0);
+      if (!hdl)
+      {
+         fprintf(stderr, "Cannot open sndio device\n");
+         quit(1);
+      }
+
+      sio_initpar(&par);
+      par.sig = 1;
+      par.bits = 16;
+      par.rate = rate;
+      par.pchan = *channels;
+
+      if (!sio_setpar(hdl, &par) || !sio_getpar(hdl, &par) ||
+        par.sig != 1 || par.bits != 16 || par.rate != rate) {
+          fprintf(stderr, "could not set sndio parameters\n");
+          quit(1);
+      }
+      *channels = par.pchan;
+      if (!sio_start(hdl)) {
+          fprintf(stderr, "could not start sndio\n");
+          quit(1);
       }
 #elif defined HAVE_SYS_AUDIOIO_H
       audio_info_t info;
@@ -309,7 +378,7 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       if (audio_fd<0)
       {
          perror("Cannot open /dev/audio");
-         exit(1);
+         quit(1);
       }
 
       AUDIO_INITINFO(&info);
@@ -324,13 +393,13 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       if (ioctl(audio_fd, AUDIO_SETINFO, &info) < 0)
       {
          perror ("AUDIO_SETINFO");
-         exit(1);
+         quit(1);
       }
       fout = fdopen(audio_fd, "w");
       if(!fout)
       {
         perror("Cannot open output");
-        exit(1);
+        quit(1);
       }
 #elif defined WIN32 || defined _WIN32
       {
@@ -338,12 +407,12 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
          if (Set_WIN_Params (INVALID_FILEDESC, rate, SAMPLE_SIZE, opus_channels))
          {
             fprintf (stderr, "Can't access %s\n", "WAVE OUT");
-            exit(1);
+            quit(1);
          }
       }
 #else
       fprintf (stderr, "No soundcard support\n");
-      exit(1);
+      quit(1);
 #endif
    } else {
       if (strcmp(outFile,"-")==0)
@@ -355,14 +424,21 @@ FILE *out_file_open(char *outFile, int rate, int *channels)
       }
       else
       {
-         fout = fopen(outFile, "wb");
+         fout = fopen_utf8(outFile, "wb");
          if (!fout)
          {
             perror(outFile);
-            exit(1);
+            quit(1);
+         }
+      }
+      if (*wav_format)
+      {
+         *wav_format = write_wav_header(fout, rate, mapping_family, *channels);
+         if (*wav_format < 0)
+         {
+            fprintf (stderr, "Error writing WAV header.\n");
+            quit(1);
          }
-         if (strcmp(outFile+strlen(outFile)-4,".wav")==0 || strcmp(outFile+strlen(outFile)-4,".WAV")==0)
-            write_wav_header(fout, rate, *channels);
       }
    }
    return fout;
@@ -381,35 +457,40 @@ void usage(void)
    printf ("output_file can be:\n");
    printf ("  filename.wav         Wav file\n");
    printf ("  filename.*           Raw PCM file (any extension other than .wav)\n");
-   printf ("  -                    stdout\n");
+   printf ("  -                    stdout (raw; unless --force-wav)\n");
    printf ("  (nothing)            Will be played to soundcard\n");
    printf ("\n");
    printf ("Options:\n");
-   printf (" --mono                Force decoding in mono\n");
-   printf (" --stereo              Force decoding in stereo\n");
    printf (" --rate n              Force decoding at sampling rate n Hz\n");
+   printf (" --gain n              Manually adjust gain by n.nn dB (0 default)\n");
    printf (" --no-dither           Do not dither 16-bit output\n");
+   printf (" --force-wav           Force wav header on output\n");
    printf (" --packet-loss n       Simulate n %% random packet loss\n");
    printf (" --save-range file     Saves check values for every frame to a file\n");
    printf (" -h, --help            This help\n");
-   printf (" -v, --version         Version information\n");
+   printf (" -V, --version         Version information\n");
    printf (" --quiet               Quiet mode\n");
    printf ("\n");
 }
 
 void version(void)
 {
-   printf("opusdec %s %s (based on %s)\n",PACKAGE,VERSION,opus_get_version_string());
+   printf("opusdec %s %s (using %s)\n",PACKAGE,VERSION,opus_get_version_string());
    printf("Copyright (C) 2008-2012 Xiph.Org Foundation\n");
 }
 
 void version_short(void)
 {
-   printf("opusdec %s %s (based on %s)\n",PACKAGE,VERSION,opus_get_version_string());
+   printf("opusdec %s %s (using %s)\n",PACKAGE,VERSION,opus_get_version_string());
    printf("Copyright (C) 2008-2012 Xiph.Org Foundation\n");
 }
 
-static OpusMSDecoder *process_header(ogg_packet *op, opus_int32 *rate, int *channels, int *preskip, float *gain, int *streams, int quiet)
+/*Process an Opus header and setup the opus decoder based on it.
+  It takes several pointers for header values which are needed
+  elsewhere in the code.*/
+static OpusMSDecoder *process_header(ogg_packet *op, opus_int32 *rate,
+       int *mapping_family, int *channels, int *preskip, float *gain,
+       float manual_gain, int *streams, int wav_format, int quiet)
 {
    int err;
    OpusMSDecoder *st;
@@ -421,7 +502,9 @@ static OpusMSDecoder *process_header(ogg_packet *op, opus_int32 *rate, int *chan
       return NULL;
    }
 
+   *mapping_family = header.channel_mapping;
    *channels = header.channels;
+   if(wav_format)adjust_wav_mapping(*mapping_family, *channels, header.stream_map);
 
    if(!*rate)*rate=header.input_sample_rate;
    /*If the rate is unspecified we decode to 48000*/
@@ -445,15 +528,35 @@ static OpusMSDecoder *process_header(ogg_packet *op, opus_int32 *rate, int *chan
 
    *streams=header.nb_streams;
 
-   *gain = pow(10., header.gain/5120.);
+   if(header.gain!=0 || manual_gain!=0)
+   {
+      /*Gain API added in a newer libopus version, if we don't have it
+        we apply the gain ourselves. We also add in a user provided
+        manual gain at the same time.*/
+      int gainadj = (int)(manual_gain*256.)+header.gain;
+#ifdef OPUS_SET_GAIN
+      err=opus_multistream_decoder_ctl(st,OPUS_SET_GAIN(gainadj));
+      if(err==OPUS_UNIMPLEMENTED)
+      {
+#endif
+         *gain = pow(10., gainadj/5120.);
+#ifdef OPUS_SET_GAIN
+      } else if (err!=OPUS_OK)
+      {
+         fprintf (stderr, "Error setting gain: %s\n", opus_strerror(err));
+         return NULL;
+      }
+#endif
+   }
 
    if (!quiet)
    {
-      fprintf(stderr, "Decoding %d Hz %saudio", *rate, *rate!=(int)header.input_sample_rate?"(forced) ":"");
-      fprintf(stderr, " (%d channel%s)",*channels,*channels>1?"s":"");
+      fprintf(stderr, "Decoding to %d Hz (%d channel%s)", *rate,
+        *channels, *channels>1?"s":"");
       if(header.version!=1)fprintf(stderr, ", Header v%d",header.version);
       fprintf(stderr, "\n");
-      if (header.gain!=0)printf("Playback gain: %f (%f dB)\n", *gain, header.gain/256.);
+      if (header.gain!=0)fprintf(stderr,"Playback gain: %f dB\n", header.gain/256.);
+      if (manual_gain!=0)fprintf(stderr,"Manual gain: %f dB\n", manual_gain);
    }
 
    return st;
@@ -465,9 +568,11 @@ opus_int64 audio_write(float *pcm, int channels, int frame_size, FILE *fout, Spe
    opus_int64 sampout=0;
    int i,ret,tmp_skip;
    unsigned out_len;
-   short out[MAX_FRAME_SIZE*channels];
-   float buf[MAX_FRAME_SIZE*channels];
+   short *out;
+   float *buf;
    float *output;
+   out=alloca(sizeof(short)*MAX_FRAME_SIZE*channels);
+   buf=alloca(sizeof(float)*MAX_FRAME_SIZE*channels);
    maxout=maxout<0?0:maxout;
    do {
      if (skip){
@@ -477,8 +582,8 @@ opus_int64 audio_write(float *pcm, int channels, int frame_size, FILE *fout, Spe
        tmp_skip = 0;
      }
      if (resampler){
-       output=buf;
        unsigned in_len;
+       output=buf;
        in_len = frame_size-tmp_skip;
        out_len = 1024<maxout?1024:maxout;
        speex_resampler_process_interleaved_float(resampler, pcm+channels*tmp_skip, &in_len, buf, &out_len);
@@ -495,7 +600,7 @@ opus_int64 audio_write(float *pcm, int channels, int frame_size, FILE *fout, Spe
        shape_dither_toshort(shapemem,out,output,out_len,channels);
      }else{
        for (i=0;i<(int)out_len*channels;i++)
-         out[i]=(short)lrintf(fmax(-32768,fmin(output[i]*32768.f,32767)));
+         out[i]=(short)float2int(fmaxf(-32768,fminf(output[i]*32768.f,32767)));
      }
      if ((le_short(1)!=1)&&file){
        for (i=0;i<(int)out_len*channels;i++)
@@ -510,6 +615,12 @@ opus_int64 audio_write(float *pcm, int channels, int frame_size, FILE *fout, Spe
          if(ret>0)ret/=sizeof(short)*channels;
          else fprintf(stderr, "Error playing audio.\n");
        }else
+#elif defined HAVE_LIBSNDIO
+       if(!file){
+         ret=sio_write (hdl, out, sizeof(short) * channels * (out_len<maxout?out_len:maxout));
+         if(ret>0)ret/=sizeof(short)*channels;
+         else fprintf(stderr, "Error playing audio.\n");
+       }else
 #endif
          ret=fwrite(out, 2*channels, out_len<maxout?out_len:maxout, fout);
        sampout+=ret;
@@ -532,6 +643,7 @@ int main(int argc, char **argv)
    int total_links=0;
    int stream_init = 0;
    int quiet = 0;
+   int forcewav = 0;
    ogg_int64_t page_granule=0;
    ogg_int64_t link_out=0;
    struct option long_options[] =
@@ -541,9 +653,9 @@ int main(int argc, char **argv)
       {"version", no_argument, NULL, 0},
       {"version-short", no_argument, NULL, 0},
       {"rate", required_argument, NULL, 0},
-      {"mono", no_argument, NULL, 0},
-      {"stereo", no_argument, NULL, 0},
+      {"gain", required_argument, NULL, 0},
       {"no-dither", no_argument, NULL, 0},
+      {"force-wav", no_argument, NULL, 0},
       {"packet-loss", required_argument, NULL, 0},
       {"save-range", required_argument, NULL, 0},
       {0, 0, 0, 0}
@@ -555,8 +667,11 @@ int main(int argc, char **argv)
    int close_in=0;
    int eos=0;
    ogg_int64_t audio_size=0;
+   double last_coded_seconds=0;
    float loss_percent=-1;
+   float manual_gain=0;
    int channels=-1;
+   int mapping_family;
    int rate=0;
    int wav_format=0;
    int preskip=0;
@@ -568,6 +683,24 @@ int main(int argc, char **argv)
    SpeexResamplerState *resampler=NULL;
    float gain=1;
    int streams=0;
+   size_t last_spin=0;
+#ifdef WIN_UNICODE
+   int argc_utf8;
+   char **argv_utf8;
+#endif
+
+   if(query_cpu_support()){
+     fprintf(stderr,"\n\n** WARNING: This program with compiled with SSE%s\n",query_cpu_support()>1?"2":"");
+     fprintf(stderr,"            but this CPU claims to lack these instructions. **\n\n");
+   }
+
+#ifdef WIN_UNICODE
+   (void)argc;
+   (void)argv;
+
+   init_console_utf8();
+   init_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+#endif
 
    output=0;
    shapemem.a_buf=0;
@@ -578,7 +711,7 @@ int main(int argc, char **argv)
    /*Process options*/
    while(1)
    {
-      c = getopt_long (argc, argv, "hvV",
+      c = getopt_long (argc_utf8, argv_utf8, "hV",
                        long_options, &option_index);
       if (c==-1)
          break;
@@ -589,37 +722,37 @@ int main(int argc, char **argv)
          if (strcmp(long_options[option_index].name,"help")==0)
          {
             usage();
-            exit(0);
+            quit(0);
          } else if (strcmp(long_options[option_index].name,"quiet")==0)
          {
             quiet = 1;
          } else if (strcmp(long_options[option_index].name,"version")==0)
          {
             version();
-            exit(0);
+            quit(0);
          } else if (strcmp(long_options[option_index].name,"version-short")==0)
          {
             version_short();
-            exit(0);
-         } else if (strcmp(long_options[option_index].name,"mono")==0)
-         {
-            channels=1;
-         } else if (strcmp(long_options[option_index].name,"stereo")==0)
-         {
-            channels=2;
+            quit(0);
          } else if (strcmp(long_options[option_index].name,"no-dither")==0)
          {
             dither=0;
+         } else if (strcmp(long_options[option_index].name,"force-wav")==0)
+         {
+            forcewav=1;
          } else if (strcmp(long_options[option_index].name,"rate")==0)
          {
             rate=atoi (optarg);
-        }else if(strcmp(long_options[option_index].name,"save-range")==0){
-          frange=fopen(optarg,"w");
+         } else if (strcmp(long_options[option_index].name,"gain")==0)
+         {
+            manual_gain=atof (optarg);
+         }else if(strcmp(long_options[option_index].name,"save-range")==0){
+          frange=fopen_utf8(optarg,"w");
           if(frange==NULL){
             perror(optarg);
             fprintf(stderr,"Could not open save-range file: %s\n",optarg);
             fprintf(stderr,"Must provide a writable file name.\n");
-            exit(1);
+            quit(1);
           }
          } else if (strcmp(long_options[option_index].name,"packet-loss")==0)
          {
@@ -628,32 +761,50 @@ int main(int argc, char **argv)
          break;
       case 'h':
          usage();
-         exit(0);
+         quit(0);
          break;
-      case 'v':
+      case 'V':
          version();
-         exit(0);
+         quit(0);
          break;
       case '?':
          usage();
-         exit(1);
+         quit(1);
          break;
       }
    }
-   if (argc-optind!=2 && argc-optind!=1)
+   if (argc_utf8-optind!=2 && argc_utf8-optind!=1)
    {
       usage();
-      exit(1);
+      quit(1);
+   }
+   inFile=argv_utf8[optind];
+
+   /*Output to a file or playback?*/
+   if (argc_utf8-optind==2){
+     /*If we're outputting to a file, should we apply a wav header?*/
+     int i;
+     char *ext;
+     outFile=argv_utf8[optind+1];
+     ext=".wav";
+     i=strlen(outFile)-4;
+     wav_format=i>=0;
+     while(wav_format&&ext&&outFile[i]) {
+       wav_format&=tolower(outFile[i++])==*ext++;
+     }
+     wav_format|=forcewav;
+   }else {
+     outFile="";
+     wav_format=0;
+     /*If playing to audio out, default the rate to 48000
+       instead of the original rate. The original rate is
+       only important for minimizing surprise about the rate
+       of output files and preserving length, which aren't
+       relevant for playback. Many audio devices sound
+       better at 48kHz and not resampling also saves CPU.*/
+     if(rate==0)rate=48000;
    }
-   inFile=argv[optind];
 
-   if (argc-optind==2)
-      outFile=argv[optind+1];
-   else
-      outFile = "";
-   wav_format = strlen(outFile)>=4 && (
-                                       strcmp(outFile+strlen(outFile)-4,".wav")==0
-                                       || strcmp(outFile+strlen(outFile)-4,".WAV")==0);
    /*Open input file*/
    if (strcmp(inFile, "-")==0)
    {
@@ -664,21 +815,27 @@ int main(int argc, char **argv)
    }
    else
    {
-      fin = fopen(inFile, "rb");
+      fin = fopen_utf8(inFile, "rb");
       if (!fin)
       {
          perror(inFile);
-         exit(1);
+         quit(1);
       }
       close_in=1;
    }
 
-
-   /*Init Ogg data struct*/
+   /* .opus files use the Ogg container to provide framing and timekeeping.
+    * http://tools.ietf.org/html/draft-terriberry-oggopus
+    * The easiest way to decode the Ogg container is to use libogg, so
+    *  thats what we do here.
+    * Using libogg is fairly straight forward-- you take your stream of bytes
+    *  and feed them to ogg_sync_ and it periodically returns Ogg pages, you
+    *  check if the pages belong to the stream you're decoding then you give
+    *  them to libogg and it gives you packets. You decode the packets. The
+    *  pages also provide timing information.*/
    ogg_sync_init(&oy);
 
    /*Main decoding loop*/
-
    while (1)
    {
       char *data;
@@ -706,6 +863,8 @@ int main(int argc, char **argv)
          /*Extract all available packets*/
          while (ogg_stream_packetout(&os, &op) == 1)
          {
+            /*OggOpus streams are identified by a magic string in the initial
+              stream header.*/
             if (op.b_o_s && op.bytes>=8 && !memcmp(op.packet, "OpusHead", 8)) {
                if(!has_opus_stream)
                {
@@ -721,13 +880,28 @@ int main(int argc, char **argv)
             }
             if (!has_opus_stream || os.serialno != opus_serialno)
                break;
-            /*If first packet, process as OPUS header*/
+            /*If first packet in a logical stream, process the Opus header*/
             if (packet_count==0)
             {
-               st = process_header(&op, &rate, &channels, &preskip, &gain, &streams, quiet);
+               st = process_header(&op, &rate, &mapping_family, &channels, &preskip, &gain, manual_gain, &streams, wav_format, quiet);
                if (!st)
-                  exit(1);
+                  quit(1);
+
+               if(ogg_stream_packetout(&os, &op)!=0 || og.header[og.header_len-1]==255)
+               {
+                  /*The format specifies that the initial header and tags packets are on their
+                    own pages. To aid implementors in discovering that their files are wrong
+                    we reject them explicitly here. In some player designs files like this would
+                    fail even without an explicit test.*/
+                  fprintf(stderr, "Extra packets on initial header page. Invalid stream.\n");
+                  quit(1);
+               }
+
+               /*Remember how many samples at the front we were told to skip
+                 so that we can adjust the timestamp counting.*/
                gran_offset=preskip;
+
+               /*Setup the memory for the dithered output*/
                if(!shapemem.a_buf)
                {
                   shapemem.a_buf=calloc(channels,sizeof(float)*4);
@@ -735,6 +909,11 @@ int main(int argc, char **argv)
                   shapemem.fs=rate;
                }
                if(!output)output=malloc(sizeof(float)*MAX_FRAME_SIZE*channels);
+
+               /*Normal players should just play at 48000 or their maximum rate,
+                 as described in the OggOpus spec.  But for commandline tools
+                 like opusdec it can be desirable to exactly preserve the original
+                 sampling rate and duration, so we have a resampler here.*/
                if (rate != 48000)
                {
                   int err;
@@ -743,13 +922,20 @@ int main(int argc, char **argv)
                      fprintf(stderr, "resampler error: %s\n", speex_resampler_strerror(err));
                   speex_resampler_skip_zeros(resampler);
                }
-               if(!fout)fout=out_file_open(outFile, rate, &channels);
+               if(!fout)fout=out_file_open(outFile, &wav_format, rate, mapping_family, &channels);
             } else if (packet_count==1)
             {
                if (!quiet)
                   print_comments((char*)op.packet, op.bytes);
+               if(ogg_stream_packetout(&os, &op)!=0 || og.header[og.header_len-1]==255)
+               {
+                  fprintf(stderr, "Extra packets on initial tags page. Invalid stream.\n");
+                  quit(1);
+               }
             } else {
+               int ret;
                opus_int64 maxout;
+               opus_int64 outsamp;
                int lost=0;
                if (loss_percent>0 && 100*((float)rand())/RAND_MAX<loss_percent)
                   lost=1;
@@ -757,14 +943,45 @@ int main(int argc, char **argv)
                /*End of stream condition*/
                if (op.e_o_s && os.serialno == opus_serialno)eos=1; /* don't care for anything except opus eos */
 
-               int ret;
-               opus_int64 outsamp;
-               /*Decode frame*/
-               if (!lost)
+               /*Are we simulating loss for this packet?*/
+               if (!lost){
+                  /*Decode Opus packet*/
                   ret = opus_multistream_decode_float(st, (unsigned char*)op.packet, op.bytes, output, MAX_FRAME_SIZE, 0);
-               else
-                  ret = opus_multistream_decode_float(st, NULL, 0, output, MAX_FRAME_SIZE, 0);
+               } else {
+                  /*Extract the original duration.
+                    Normally you wouldn't have it for a lost packet, but normally the
+                    transports used on lossy channels will effectively tell you.
+                    This avoids opusdec squaking when the decoded samples and
+                    granpos mismatches.*/
+                  opus_int32 lost_size;
+                  lost_size = MAX_FRAME_SIZE;
+                  if(op.bytes>0){
+                    opus_int32 spp;
+                    spp=opus_packet_get_nb_frames(op.packet,op.bytes);
+                    if(spp>0){
+                      spp*=opus_packet_get_samples_per_frame(op.packet,48000/*decoding_rate*/);
+                      if(spp>0)lost_size=spp;
+                    }
+                  }
+                  /*Invoke packet loss concealment.*/
+                  ret = opus_multistream_decode_float(st, NULL, 0, output, lost_size, 0);
+               }
 
+               if(!quiet){
+                  /*Display a progress spinner while decoding.*/
+                  static const char spinner[]="|/-\\";
+                  double coded_seconds = (double)audio_size/(channels*rate*sizeof(short));
+                  if(coded_seconds>=last_coded_seconds+1){
+                     fprintf(stderr,"\r[%c] %02d:%02d:%02d", spinner[last_spin&3],
+                             (int)(coded_seconds/3600),(int)(coded_seconds/60)%60,
+                             (int)(coded_seconds)%60);
+                     fflush(stderr);
+                     last_spin++;
+                     last_coded_seconds=coded_seconds;
+                  }
+               }
+
+               /*If the decoder returned less than zero, we have an error.*/
                if (ret<0)
                {
                   fprintf (stderr, "Decoding error: %s\n", opus_strerror(ret));
@@ -772,9 +989,10 @@ int main(int argc, char **argv)
                }
                frame_size = ret;
 
+               /*If we're collecting --save-range debugging data, collect it now.*/
                if(frange!=NULL){
                  OpusDecoder *od;
-                 opus_uint32 rngs[streams];
+                 opus_uint32 rngs[256];
                  for(i=0;i<streams;i++){
                    ret=opus_multistream_decoder_ctl(st,OPUS_MULTISTREAM_GET_DECODER_STATE(i,&od));
                    ret=opus_decoder_ctl(od,OPUS_GET_FINAL_RANGE(&rngs[i]));
@@ -783,10 +1001,16 @@ int main(int argc, char **argv)
                             rngs,streams);
                }
 
-               /* Apply header gain */
-               for (i=0;i<frame_size*channels;i++)
-                  output[i] *= gain;
+               /*Apply header gain, if we're not using an opus library new
+                 enough to do this internally.*/
+               if (gain!=0){
+                 for (i=0;i<frame_size*channels;i++)
+                    output[i] *= gain;
+               }
 
+               /*This handles making sure that our output duration respects
+                 the final end-trim by not letting the output sample count
+                 get ahead of the granpos indicated value.*/
                maxout=((page_granule-gran_offset)*rate/48000)-link_out;
                outsamp=audio_write(output, channels, frame_size, fout, resampler, &preskip, dither?&shapemem:0, strlen(outFile)!=0,0>maxout?0:maxout);
                link_out+=outsamp;
@@ -794,7 +1018,7 @@ int main(int argc, char **argv)
             }
             packet_count++;
          }
-         /* Drain the resampler */
+         /*We're done, drain the resampler if we were using it.*/
          if(eos && resampler)
          {
             float *zeros;
@@ -807,7 +1031,7 @@ int main(int argc, char **argv)
                int tmp = drain;
                if (tmp > 100)
                   tmp = 100;
-               outsamp=audio_write(zeros, channels, tmp, fout, resampler, NULL, &shapemem, strlen(outFile)==0, ((page_granule-gran_offset)*rate/48000)-link_out);
+               outsamp=audio_write(zeros, channels, tmp, fout, resampler, NULL, &shapemem, strlen(outFile)!=0, ((page_granule-gran_offset)*rate/48000)-link_out);
                link_out+=outsamp;
                audio_size+=sizeof(short)*outsamp*channels;
                drain -= tmp;
@@ -823,18 +1047,24 @@ int main(int argc, char **argv)
             st=NULL;
          }
       }
-      if (feof(fin))
+      if (feof(fin)) {
+         if(!quiet) {
+           fprintf(stderr, "\rDecoding complete.        \n");
+           fflush(stderr);
+         }
          break;
+      }
    }
 
-   if (fout && wav_format && audio_size<0x7FFFFFFF)
+   /*If we were writing wav, go set the duration.*/
+   if (strlen(outFile)!=0 && fout && wav_format>0 && audio_size<0x7FFFFFFF)
    {
       if (fseek(fout,4,SEEK_SET)==0)
       {
          int tmp;
-         tmp = le_int(audio_size+36);
+         tmp = le_int(audio_size+20+wav_format);
          if(fwrite(&tmp,4,1,fout)!=1)fprintf(stderr,"Error writing end length.\n");
-         if (fseek(fout,32,SEEK_CUR)==0)
+         if (fseek(fout,16+wav_format,SEEK_CUR)==0)
          {
             tmp = le_int(audio_size);
             if(fwrite(&tmp,4,1,fout)!=1)fprintf(stderr,"Error writing header length.\n");
@@ -843,10 +1073,11 @@ int main(int argc, char **argv)
             fprintf (stderr, "First seek worked, second didn't\n");
          }
       } else {
-         fprintf (stderr, "Cannot seek on wave file, size will be incorrect\n");
+         fprintf (stderr, "Cannot seek on wav file output, wav size chunk will be incorrect\n");
       }
    }
 
+   /*Did we make it to the end without recovering ANY opus logical streams?*/
    if(!total_links)fprintf (stderr, "This doesn't look like a Opus file\n");
 
    if (stream_init)
@@ -870,5 +1101,10 @@ int main(int argc, char **argv)
    if (fout != NULL)
       fclose(fout);
 
+#ifdef WIN_UNICODE
+   free_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+   uninit_console_utf8();
+#endif
+
    return 0;
 }
