@@ -36,18 +36,27 @@
 
 #include <stdlib.h>
 #include <string.h>
+#if (!defined WIN32 && !defined _WIN32) || defined(__MINGW32__)
 #include <unistd.h>
-#include <sys/time.h>
+#include <time.h>
+#endif
 #include <math.h>
 
 #ifdef _MSC_VER
 #define snprintf _snprintf
 #endif
 
-#ifdef _WIN32
-#define I64FORMAT "I64d"
+#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
+# include "unicode_support.h"
+/* We need the following two to set stdout to binary */
+# include <io.h>
+# include <fcntl.h>
+# define I64FORMAT "I64d"
 #else
-#define I64FORMAT "lld"
+# define I64FORMAT "lld"
+# define fopen_utf8(_x,_y) fopen((_x),(_y))
+# define argc_utf8 argc
+# define argv_utf8 argv
 #endif
 
 #include <opus.h>
@@ -58,12 +67,7 @@
 #include "opus_header.h"
 #include "opusenc.h"
 #include "diag_range.h"
-
-#if defined WIN32 || defined _WIN32
-/* We need the following two to set stdout to binary */
-#include <io.h>
-#include <fcntl.h>
-#endif
+#include "cpusupport.h"
 
 #ifdef VALGRIND
 #include <valgrind/memcheck.h>
@@ -109,7 +113,7 @@ void usage(void)
   printf("Encodes input_file using Opus. It can read the WAV, AIFF, or raw files.\n");
   printf("\nGeneral options:\n");
   printf(" -h, --help         This help\n");
-  printf(" -v, --version      Version information\n");
+  printf(" -V, --version      Version information\n");
   printf(" --quiet            Quiet mode\n");
   printf("\n");
   printf("input_file can be:\n");
@@ -120,13 +124,11 @@ void usage(void)
   printf("  filename.opus     compressed file\n");
   printf("  -                 stdout\n");
   printf("\nEncoding options:\n");
-  printf(" --speech           Optimize for speech\n");
-  printf(" --music            Optimize for music\n");
   printf(" --bitrate n.nnn    Encoding bitrate in kbit/sec (6-256 per channel)\n");
   printf(" --vbr              Use variable bitrate encoding (default)\n");
   printf(" --cvbr             Use constrained variable bitrate encoding\n");
   printf(" --hard-cbr         Use hard constant bitrate encoding\n");
-  printf(" --comp n           Encoding complexity (0-10, default: 10)\n");
+  printf(" --comp n           Encoding complexity (0-10, default: 10 (slowest))\n");
   printf(" --framesize n      Maximum frame size in milliseconds\n");
   printf("                      (2.5, 5, 10, 20, 40, 60, default: 20)\n");
   printf(" --expect-loss      Percentage packet loss to expect (default: 0)\n");
@@ -139,7 +141,6 @@ void usage(void)
   printf(" --set-ctl-int x=y  Pass the encoder control x with value y (advanced)\n");
   printf("                      Preface with s: to direct the ctl to multistream s\n");
   printf("                      This may be used multiple times\n");
-  printf(" --uncoupled        Use one mono stream per channel\n");
   printf("\nMetadata options:\n");
   printf(" --comment          Add the given string as an extra comment\n");
   printf("                      This may be used multiple times\n");
@@ -180,6 +181,7 @@ int main(int argc, char **argv)
     {"vbr",no_argument,NULL, 0},
     {"cvbr",no_argument,NULL, 0},
     {"comp", required_argument, NULL, 0},
+    {"complexity", required_argument, NULL, 0},
     {"framesize", required_argument, NULL, 0},
     {"expect-loss", required_argument, NULL, 0},
     {"downmix-mono",no_argument,NULL, 0},
@@ -197,8 +199,6 @@ int main(int argc, char **argv)
     {"raw-endianness", required_argument, NULL, 0},
     {"ignorelength", no_argument, NULL, 0},
     {"rate", required_argument, NULL, 0},
-    {"music", no_argument, NULL, 0},
-    {"speech", no_argument, NULL, 0},
     {"version", no_argument, NULL, 0},
     {"version-short", no_argument, NULL, 0},
     {"comment", required_argument, NULL, 0},
@@ -243,8 +243,8 @@ int main(int argc, char **argv)
   opus_int32         nb_samples;
   opus_int32         peak_bytes=0;
   opus_int32         min_bytes;
-  struct timeval     start_time;
-  struct timeval     stop_time;
+  time_t             start_time;
+  time_t             stop_time;
   time_t             last_spin=0;
   int                last_spin_len=0;
   /*Settings*/
@@ -257,7 +257,6 @@ int main(int argc, char **argv)
   int                chan=2;
   int                with_hard_cbr=0;
   int                with_cvbr=0;
-  int                signal=OPUS_AUTO;
   int                expect_loss=0;
   int                complexity=10;
   int                downmix=0;
@@ -268,6 +267,22 @@ int main(int argc, char **argv)
   opus_int32         lookahead=0;
   unsigned char      mapping[256];
   int                force_narrow=0;
+#ifdef WIN_UNICODE
+   int argc_utf8;
+   char **argv_utf8;
+#endif
+
+   if(query_cpu_support()){
+     fprintf(stderr,"\n\n** WARNING: This program was compiled with SSE%s\n",query_cpu_support()>1?"2":"");
+     fprintf(stderr,"            but this CPU claims to lack these instructions. **\n\n");
+   }
+
+#ifdef WIN_UNICODE
+   (void)argc;
+   (void)argv;
+
+   init_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+#endif
 
   opt_ctls_ctlval=NULL;
   frange=NULL;
@@ -286,13 +301,13 @@ int main(int argc, char **argv)
   /*Vendor string should just be the encoder library,
     the ENCODER comment specifies the tool used.*/
   comment_init(&comments, &comments_length, opus_version);
-  snprintf(ENCODER_string, sizeof(ENCODER_string), "opusenc from %s %s\n",PACKAGE,VERSION);
+  snprintf(ENCODER_string, sizeof(ENCODER_string), "opusenc from %s %s",PACKAGE,VERSION);
   comment_add(&comments, &comments_length, "ENCODER=", ENCODER_string);
 
   /*Process command-line options*/
   while(1){
     int c;
-    c=getopt_long(argc, argv, "hv",
+    c=getopt_long(argc_utf8, argv_utf8, "hV",
                   long_options, &option_index);
     if(c==-1)
        break;
@@ -342,10 +357,6 @@ int main(int argc, char **argv)
         }else if(strcmp(long_options[option_index].name,"raw-endianness")==0){
           inopt.rawmode=1;
           inopt.endianness=atoi(optarg);
-        }else if(strcmp(long_options[option_index].name,"music")==0){
-          signal=OPUS_SIGNAL_MUSIC;
-        }else if(strcmp(long_options[option_index].name,"speech")==0){
-          signal=OPUS_SIGNAL_VOICE;
         }else if(strcmp(long_options[option_index].name,"downmix-mono")==0){
           downmix=1;
         }else if(strcmp(long_options[option_index].name,"downmix-stereo")==0){
@@ -359,7 +370,8 @@ int main(int argc, char **argv)
             fprintf(stderr,"Expected loss is a percent and must be 0-100.\n");
             exit(1);
           }
-        }else if(strcmp(long_options[option_index].name,"comp")==0){
+        }else if(strcmp(long_options[option_index].name,"comp")==0 ||
+                 strcmp(long_options[option_index].name,"complexity")==0){
           complexity=atoi(optarg);
           if(complexity>10||complexity<0){
             fprintf(stderr,"Invalid complexity: %s\n",optarg);
@@ -412,7 +424,7 @@ int main(int argc, char **argv)
           opt_ctls_ctlval[opt_ctls*3+2]=atoi(spos+1);
           opt_ctls++;
         }else if(strcmp(long_options[option_index].name,"save-range")==0){
-          frange=fopen(optarg,"w");
+          frange=fopen_utf8(optarg,"w");
           if(frange==NULL){
             perror(optarg);
             fprintf(stderr,"Could not open save-range file: %s\n",optarg);
@@ -439,7 +451,7 @@ int main(int argc, char **argv)
         usage();
         exit(0);
         break;
-      case 'v':
+      case 'V':
         opustoolsversion(opus_version);
         exit(0);
         break;
@@ -449,12 +461,12 @@ int main(int argc, char **argv)
         break;
     }
   }
-  if(argc-optind!=2){
+  if(argc_utf8-optind!=2){
     usage();
     exit(1);
   }
-  inFile=argv[optind];
-  outFile=argv[optind+1];
+  inFile=argv_utf8[optind];
+  outFile=argv_utf8[optind+1];
 
   if(strcmp(inFile, "-")==0){
 #if defined WIN32 || defined _WIN32
@@ -464,7 +476,7 @@ int main(int argc, char **argv)
 #endif
     fin=stdin;
   }else{
-    fin=fopen(inFile, "rb");
+    fin=fopen_utf8(inFile, "rb");
     if(!fin){
       perror(inFile);
       exit(1);
@@ -506,7 +518,12 @@ int main(int argc, char **argv)
 
   /*Scale the resampler complexity, but only for 48000 output because
     the near-cutoff behavior matters a lot more at lower rates.*/
-  if(rate!=coding_rate)setup_resample(&inopt,coding_rate==48000?complexity/2:5,coding_rate);
+  if(rate!=coding_rate)setup_resample(&inopt,coding_rate==48000?(complexity+1)/2:5,coding_rate);
+
+  if(rate!=coding_rate&&complexity!=10&&!quiet){
+    fprintf(stderr,"Notice: Using resampling with complexity<10.\n");
+    fprintf(stderr,"Opusenc is fastest with 48, 24, 16, 12, or 8kHz input.\n\n");
+  }
 
   /*OggOpus headers*/ /*FIXME: broke forcemono*/
   header.channels=chan;
@@ -548,13 +565,7 @@ int main(int argc, char **argv)
   st=opus_multistream_encoder_create(coding_rate, chan, header.nb_streams, header.nb_coupled,
      mapping, frame_size<480/(48000/coding_rate)?OPUS_APPLICATION_RESTRICTED_LOWDELAY:OPUS_APPLICATION_AUDIO, &ret);
   if(ret!=OPUS_OK){
-    fprintf(stderr, "Cannot create encoder: %s\n", opus_strerror(ret));
-    exit(1);
-  }
-
-  ret=opus_multistream_encoder_ctl(st, OPUS_SET_SIGNAL(signal));
-  if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_SET_SIGNAL returned: %s\n",opus_strerror(ret));
+    fprintf(stderr, "Error cannot create encoder: %s\n", opus_strerror(ret));
     exit(1);
   }
 
@@ -565,14 +576,18 @@ int main(int argc, char **argv)
         opus_multistream_encoder_ctl(st,OPUS_MULTISTREAM_GET_ENCODER_STATE(i,&oe));
         ret=opus_encoder_ctl(oe, OPUS_SET_MAX_BANDWIDTH(OPUS_BANDWIDTH_NARROWBAND));
         if(ret!=OPUS_OK){
-          fprintf(stderr,"OPUS_SET_MAX_BANDWIDTH on stream %d returned: %s\n",i,opus_strerror(ret));
+          fprintf(stderr,"Error OPUS_SET_MAX_BANDWIDTH on stream %d returned: %s\n",i,opus_strerror(ret));
           exit(1);
         }
       }
     }
   }
 
-  bitrate=bitrate>0?bitrate:64000*header.nb_streams+32000*header.nb_coupled;
+  if(bitrate<0){
+    /*Lower default rate for sampling rates [8000-44100) by a factor of (rate+16k)/(64k)*/
+    bitrate=((64000*header.nb_streams+32000*header.nb_coupled)*
+             (IMIN(48,IMAX(8,((rate<44100?rate:48000)+1000)/1000))+16)+32)>>6;
+  }
 
   if(bitrate>(1024000*chan)||bitrate<500){
     fprintf(stderr,"Error: Bitrate %d bits/sec is insane.\nDid you mistake bits for kilobits?\n",bitrate);
@@ -583,43 +598,50 @@ int main(int argc, char **argv)
 
   ret=opus_multistream_encoder_ctl(st, OPUS_SET_BITRATE(bitrate));
   if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_SET_BITRATE returned: %s\n",opus_strerror(ret));
+    fprintf(stderr,"Error OPUS_SET_BITRATE returned: %s\n",opus_strerror(ret));
     exit(1);
   }
 
   ret=opus_multistream_encoder_ctl(st, OPUS_SET_VBR(!with_hard_cbr));
   if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_SET_VBR returned: %s\n",opus_strerror(ret));
+    fprintf(stderr,"Error OPUS_SET_VBR returned: %s\n",opus_strerror(ret));
     exit(1);
   }
 
   if(!with_hard_cbr){
     ret=opus_multistream_encoder_ctl(st, OPUS_SET_VBR_CONSTRAINT(with_cvbr));
     if(ret!=OPUS_OK){
-      fprintf(stderr,"OPUS_SET_VBR_CONSTRAINT returned: %s\n",opus_strerror(ret));
+      fprintf(stderr,"Error OPUS_SET_VBR_CONSTRAINT returned: %s\n",opus_strerror(ret));
       exit(1);
     }
   }
 
   ret=opus_multistream_encoder_ctl(st, OPUS_SET_COMPLEXITY(complexity));
   if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_SET_COMPLEXITY returned: %s\n",opus_strerror(ret));
+    fprintf(stderr,"Error OPUS_SET_COMPLEXITY returned: %s\n",opus_strerror(ret));
     exit(1);
   }
 
   ret=opus_multistream_encoder_ctl(st, OPUS_SET_PACKET_LOSS_PERC(expect_loss));
   if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_SET_PACKET_LOSS_PERC returned: %s\n",opus_strerror(ret));
+    fprintf(stderr,"Error OPUS_SET_PACKET_LOSS_PERC returned: %s\n",opus_strerror(ret));
     exit(1);
   }
 
+#ifdef OPUS_SET_LSB_DEPTH
+  ret=opus_multistream_encoder_ctl(st, OPUS_SET_LSB_DEPTH(IMAX(8,IMIN(24,inopt.samplesize))));
+  if(ret!=OPUS_OK){
+    fprintf(stderr,"Warning OPUS_SET_LSB_DEPTH returned: %s\n",opus_strerror(ret));
+  }
+#endif
+
   /*This should be the last set of CTLs, except the lookahead get, so it can override the defaults.*/
   for(i=0;i<opt_ctls;i++){
     int target=opt_ctls_ctlval[i*3];
     if(target==-1){
       ret=opus_multistream_encoder_ctl(st,opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2]);
       if(ret!=OPUS_OK){
-        fprintf(stderr,"opus_multistream_encoder_ctl(st,%d,%d) returned: %s\n",opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2],opus_strerror(ret));
+        fprintf(stderr,"Error opus_multistream_encoder_ctl(st,%d,%d) returned: %s\n",opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2],opus_strerror(ret));
         exit(1);
       }
     }else if(target<header.nb_streams){
@@ -627,11 +649,11 @@ int main(int argc, char **argv)
       opus_multistream_encoder_ctl(st,OPUS_MULTISTREAM_GET_ENCODER_STATE(i,&oe));
       ret=opus_encoder_ctl(oe, opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2]);
       if(ret!=OPUS_OK){
-        fprintf(stderr,"opus_encoder_ctl(st[%d],%d,%d) returned: %s\n",target,opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2],opus_strerror(ret));
+        fprintf(stderr,"Error opus_encoder_ctl(st[%d],%d,%d) returned: %s\n",target,opt_ctls_ctlval[i*3+1],opt_ctls_ctlval[i*3+2],opus_strerror(ret));
         exit(1);
       }
     }else{
-      fprintf(stderr,"--set-ctl-int target stream %d is higher than the maximum stream number %d.\n",target,header.nb_streams-1);
+      fprintf(stderr,"Error --set-ctl-int target stream %d is higher than the maximum stream number %d.\n",target,header.nb_streams-1);
       exit(1);
     }
   }
@@ -639,7 +661,7 @@ int main(int argc, char **argv)
   /*We do the lookahead check late so user CTLs can change it*/
   ret=opus_multistream_encoder_ctl(st, OPUS_GET_LOOKAHEAD(&lookahead));
   if(ret!=OPUS_OK){
-    fprintf(stderr,"OPUS_GET_LOOKAHEAD returned: %s\n",opus_strerror(ret));
+    fprintf(stderr,"Error OPUS_GET_LOOKAHEAD returned: %s\n",opus_strerror(ret));
     exit(1);
   }
   inopt.skip+=lookahead;
@@ -680,7 +702,7 @@ int main(int argc, char **argv)
 #endif
     fout=stdout;
   }else{
-    fout=fopen(outFile, "wb");
+    fout=fopen_utf8(outFile, "wb");
     if(!fout){
       perror(outFile);
       exit(1);
@@ -688,13 +710,12 @@ int main(int argc, char **argv)
   }
 
   /*Initialize Ogg stream struct*/
-  gettimeofday(&start_time,NULL);
-  srand(((getpid()&65535)<<15)^start_time.tv_sec^start_time.tv_usec);
+  start_time = time(NULL);
+  srand(((getpid()&65535)<<15)^start_time);
   if(ogg_stream_init(&os, rand())==-1){
     fprintf(stderr,"Error: stream init failed\n");
     exit(1);
   }
-  start_time.tv_sec=0;
 
   /*Write header*/
   {
@@ -763,7 +784,9 @@ int main(int argc, char **argv)
     }
     op.e_o_s|=eos;
 
-    if(start_time.tv_sec==0)gettimeofday(&start_time,NULL);
+    if(start_time==0){
+      start_time = time(NULL);
+    }
 
     cur_frame_size=frame_size;
 
@@ -788,8 +811,8 @@ int main(int argc, char **argv)
     min_bytes=IMIN(nbBytes,min_bytes);
 
     if(frange!=NULL){
+      opus_uint32 rngs[256];
       OpusEncoder *oe;
-      opus_uint32 rngs[header.nb_streams];
       for(i=0;i<header.nb_streams;i++){
         ret=opus_multistream_encoder_ctl(st,OPUS_MULTISTREAM_GET_ENCODER_STATE(i,&oe));
         ret=opus_encoder_ctl(oe,OPUS_GET_FINAL_RANGE(&rngs[i]));
@@ -846,7 +869,7 @@ int main(int argc, char **argv)
 
     /*If the stream is over or we're sure that the delayed flush will fire,
       go ahead and flush now to avoid adding delay.*/
-    while((op.e_o_s||(enc_granulepos+frame_size-last_granulepos>max_ogg_delay)||
+    while((op.e_o_s||(enc_granulepos+(frame_size*48000/coding_rate)-last_granulepos>max_ogg_delay)||
            (last_segments>=255))?
 #ifdef OLD_LIBOGG
     /*Libogg > 1.2.2 allows us to achieve lower overhead by
@@ -870,12 +893,11 @@ int main(int argc, char **argv)
     }
 
     if(!quiet){
-      gettimeofday(&stop_time,NULL);
-      if(stop_time.tv_sec>last_spin){
+      stop_time = time(NULL);
+      if(stop_time>last_spin){
         double estbitrate;
         double coded_seconds=nb_encoded/(double)coding_rate;
-        double wall_time=(stop_time.tv_sec-start_time.tv_sec)+
-          (stop_time.tv_usec-start_time.tv_usec)*1e-06;
+        double wall_time=(stop_time-start_time)+1e-6;
         char sbuf[55];
         static const char spinner[]="|/-\\";
         if(!with_hard_cbr){
@@ -883,6 +905,7 @@ int main(int argc, char **argv)
           estbitrate=(total_bytes*8.0/coded_seconds)*tweight+
                       bitrate*(1.-tweight);
         }else estbitrate=nbBytes*8*((double)coding_rate/frame_size);
+        fprintf(stderr,"\r");
         for(i=0;i<last_spin_len;i++)fprintf(stderr," ");
         if(inopt.total_samples_per_channel>0 && inopt.total_samples_per_channel<nb_encoded){
           snprintf(sbuf,54,"\r[%c] %02d%% ",spinner[last_spin&3],
@@ -892,27 +915,27 @@ int main(int argc, char **argv)
         }
         last_spin_len=strlen(sbuf);
         snprintf(sbuf+last_spin_len,54-last_spin_len,
-          "%02d:%02d:%02d.%02d %4.3gx realtime, %5.4gkbit/s\r",
+          "%02d:%02d:%02d.%02d %4.3gx realtime, %5.4gkbit/s",
           (int)(coded_seconds/3600),(int)(coded_seconds/60)%60,
           (int)(coded_seconds)%60,(int)(coded_seconds*100)%100,
           coded_seconds/wall_time,
           estbitrate/1000.);
         fprintf(stderr,"%s",sbuf);
+        fflush(stderr);
         last_spin_len=strlen(sbuf);
-        last_spin=stop_time.tv_sec;
+        last_spin=stop_time;
       }
     }
   }
-  gettimeofday(&stop_time,NULL);
+  stop_time = time(NULL);
 
   for(i=0;i<last_spin_len;i++)fprintf(stderr," ");
   if(last_spin_len)fprintf(stderr,"\r");
 
   if(!quiet){
     double coded_seconds=nb_encoded/(double)coding_rate;
-    double wall_time=(stop_time.tv_sec-start_time.tv_sec)+
-      (stop_time.tv_usec-start_time.tv_usec)*1e-06;
-    fprintf(stderr,"Encoding complete\n");
+    double wall_time=(stop_time-start_time)+1e-6;
+    fprintf(stderr,"Encoding complete                                    \n");
     fprintf(stderr,"-----------------------------------------------------\n");
     fprintf(stderr,"    Encoded:");
     print_time(coded_seconds);
@@ -945,6 +968,9 @@ int main(int argc, char **argv)
   if(fin)fclose(fin);
   if(fout)fclose(fout);
   if(frange)fclose(frange);
+#ifdef WIN_UNICODE
+   free_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+#endif
   return 0;
 }
 
