@@ -31,6 +31,14 @@
 #include "opus_header.h"
 #include "info_opus.h"
 
+#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
+# include "unicode_support.h"
+#else
+# define fopen_utf8(_x,_y) fopen((_x),(_y))
+# define argc_utf8 argc
+# define argv_utf8 argv
+#endif
+
 #define CHUNK 4500
 
 static int printlots = 0;
@@ -436,14 +444,14 @@ static stream_processor *find_stream_processor(stream_set *set, ogg_page *page)
    return stream;
 }
 
-static int get_next_page(FILE *f, ogg_sync_state *sync, ogg_page *page,
+static int get_next_page(FILE *f, ogg_sync_state *ogsync, ogg_page *page,
         ogg_int64_t *written)
 {
     int ret;
     char *buffer;
     int bytes;
 
-    while((ret = ogg_sync_pageseek(sync, page)) <= 0) {
+    while((ret = ogg_sync_pageseek(ogsync, page)) <= 0) {
         if(ret < 0) {
             /* unsynced, we jump over bytes to a possible capture - we don't need to read more just yet */
             oi_warn(_("WARNING: Hole in data (%d bytes) found at approximate offset %" I64FORMAT " bytes. Corrupted Ogg.\n"), -ret, *written);
@@ -451,13 +459,13 @@ static int get_next_page(FILE *f, ogg_sync_state *sync, ogg_page *page,
         }
 
         /* zero return, we didn't have enough data to find a whole page, read */
-        buffer = ogg_sync_buffer(sync, CHUNK);
+        buffer = ogg_sync_buffer(ogsync, CHUNK);
         bytes = fread(buffer, 1, CHUNK, f);
         if(bytes <= 0) {
-            ogg_sync_wrote(sync, 0);
+            ogg_sync_wrote(ogsync, 0);
             return 0;
         }
-        ogg_sync_wrote(sync, bytes);
+        ogg_sync_wrote(ogsync, bytes);
         *written += bytes;
     }
 
@@ -465,8 +473,8 @@ static int get_next_page(FILE *f, ogg_sync_state *sync, ogg_page *page,
 }
 
 static void process_file(char *filename) {
-    FILE *file = fopen(filename, "rb");
-    ogg_sync_state sync;
+    FILE *file = fopen_utf8(filename, "rb");
+    ogg_sync_state ogsync;
     ogg_page page;
     stream_set *processors = create_stream_set();
     int gotpage = 0;
@@ -480,9 +488,9 @@ static void process_file(char *filename) {
 
     printf(_("Processing file \"%s\"...\n\n"), filename);
 
-    ogg_sync_init(&sync);
+    ogg_sync_init(&ogsync);
 
-    while(get_next_page(file, &sync, &page, &written)) {
+    while(get_next_page(file, &ogsync, &page, &written)) {
         stream_processor *p = find_stream_processor(processors, &page);
         gotpage = 1;
 
@@ -557,7 +565,7 @@ static void process_file(char *filename) {
 
     free_stream_set(processors);
 
-    ogg_sync_clear(&sync);
+    ogg_sync_clear(&ogsync);
 
     fclose(file);
 }
@@ -583,20 +591,35 @@ static void usage(void) {
     printf (_("\t-V Output version information and exit.\n"));
 }
 
-int main(int argc, char **argv) {
+int main(int argc, char **argv)
+{
     int f, ret;
 
-    if(argc < 2) {
+#ifdef WIN_UNICODE
+    int argc_utf8;
+    char **argv_utf8;
+
+    (void)argc;
+    (void)argv;
+
+    init_console_utf8();
+    init_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+#endif
+
+    if(argc_utf8 < 2) {
         fprintf(stdout,
                 _("Usage: opusinfo [flags] file1.opus [file2.opus ... fileN.opus]\n"
                   "\n"
                   "opusinfo is a tool for printing information about Opus files\n"
                   "and for diagnosing problems with them.\n"
                   "Full help shown with \"opusinfo -h\".\n"));
+#ifdef WIN_UNICODE
+        uninit_console_utf8();
+#endif
         exit(1);
     }
 
-    while((ret = getopt(argc, argv, "hqvV")) >= 0) {
+    while((ret = getopt(argc_utf8, argv_utf8, "hqvV")) >= 0) {
         switch(ret) {
             case 'h':
                 usage();
@@ -620,7 +643,7 @@ int main(int argc, char **argv) {
     if(verbose < 0)
         printwarn = 0;
 
-    if(optind >= argc) {
+    if(optind >= argc_utf8) {
         fprintf(stderr,
                 _("No input files specified. \"opusinfo -h\" for help\n"));
         return 1;
@@ -628,12 +651,17 @@ int main(int argc, char **argv) {
 
     ret = 0;
 
-    for(f=optind; f < argc; f++) {
+    for(f=optind; f < argc_utf8; f++) {
         flawed = 0;
-        process_file(argv[f]);
+        process_file(argv_utf8[f]);
         if(flawed != 0)
             ret = flawed;
     }
 
+#ifdef WIN_UNICODE
+   free_commandline_arguments_utf8(&argc_utf8, &argv_utf8);
+   uninit_console_utf8();
+#endif
+
     return ret;
 }
